import logging
import sys
from socket import gethostname
from getpass import getuser
from pathlib import Path
import subprocess

FORMATTER = logging.Formatter("%(asctime)s — %(module)s — %(levelname)s — %(message)s")


def setup_logging(logger):
    logger.setLevel(logging.INFO)
    logger.addHandler(get_console_handler())
    # logger.propagate = False

    logger.info('Logging started')
    logger.info('Python version: %s', sys.version)
    logger.info('Script started by %s on %s', getuser(), gethostname())
    logger.info('Called with sys.argv:\n %s', ' '.join(sys.argv))
    rev_hash = get_git_revision_hash(Path(sys.argv[0]).parent)
    if rev_hash:
        logger.info('Current commit hash: %s', rev_hash)
    else:
        logger.warning("This script is not in a git repository!")

    return logger


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_git_revision_hash(git_dir):
    try:
        output = subprocess.check_output(['git', '-C', str(git_dir), 'rev-parse', 'HEAD'], stderr=subprocess.DEVNULL)
        return output.decode('utf8').rstrip()
    except subprocess.CalledProcessError:
        pass
    return None



from pathlib import Path

from dynaconf import Dynaconf


def get_configs():
    config_path = Path(__file__).parent / "configs"
    return {f.stem: f for f in config_path.iterdir() if f.is_file() and f.suffix == ".toml"}


def load_config(config_path: Path):
    settings = Dynaconf(
        envvar_prefix="DYNACONF",
        settings_files=[str(config_path)],
    )
    check_and_augment_conf(settings)
    return settings


def check_and_augment_conf(settings):
    # TODO: add decent checks
    for partition in settings.partitions:
        part = settings[partition]
        layout = part.layout
        part.layout = settings[layout]
        layout = part.layout
        layout.blocksize = (layout.pagesize + layout.oobsize) * layout.pages_per_block

        if hasattr(layout, "ecc_algorithm"):
            ecc_algorithm = layout.ecc_algorithm
            layout.ecc_algorithm = settings[ecc_algorithm] if isinstance(ecc_algorithm, str) else layout.ecc_algorithm
        else:
            layout.ecc_algorithm = None
        if not hasattr(layout, "ecc"):
            layout.ecc = None
        if not hasattr(layout, "ecc_protected_data"):
            layout.ecc_protected_data = None

        if hasattr(layout, "etfs"):
            etfs_layout = layout.etfs
            layout.etfs = settings[etfs_layout] if isinstance(etfs_layout, str) else layout.etfs

        # Set default values for optional parameters
        if not hasattr(layout, "left_shift_ecc_buf"):
            layout.left_shift_ecc_buf = 0
        if not hasattr(layout, "ecc_protected_data_reverse"):
            layout.ecc_protected_data_reverse = False
        if not hasattr(layout, "ecc_protected_data_invert"):
            layout.ecc_protected_data_invert = False
        if not hasattr(layout, "ecc_reverse"):
            layout.ecc_reverse = False
        if not hasattr(layout, "ecc_invert"):
            layout.ecc_invert = False




import stat
from pathlib import Path
import logging
import os
from time import time

from fuse import FUSE, Operations

from nandtool.nand import build_partitions
from nandtool.config import load_config


LOGGER = logging.getLogger(__name__)


class FuseNAND(Operations):
    """Fuse implementation of the ETFS file system.

    Args:
        image (Path): Path to the image.
        offset (int): Start of the ETFS partition in the image.
    """

    def __init__(self, image, mountpoint, conf):
        self.mountpoint = Path(mountpoint)
        # TODO add visual feedback that processing can take a while
        self.partitions = build_partitions(image, conf)
        LOGGER.info(f"NAND chip is now mounted at {mountpoint}")

    def getattr(self, path, fh=None):
        """Get directory with stat information.

        Args:
            path (Path): Path.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"getattr({path})")
        path = Path(path)
        st = os.lstat(self.mountpoint.parent)
        if path == Path("/"):
            return dict(
                (key, getattr(st, key))
                for key in ("st_atime", "st_ctime", "st_gid", "st_mode", "st_mtime", "st_nlink", "st_size", "st_uid")
            )
        elif path.name in self.partitions:
            part = self.partitions[path.name]
            return {
                "st_mode": 0o444 | stat.S_IFREG,
                "st_nlink": 2,
                "st_size": part.corrected_partition_size,
                "st_atime": time(),
                "st_ctime": time(),
                "st_mtime": time(),
                "st_gid": 0,
                "st_uid": 0,
            }
        return dict()

    def readdir(self, path, fh):
        """Read content from directory.

        Args:
            path (Path): Path to directory.
            fh: file handle.

        Returns:
            dict: dictionary with keys identical to the stat C structure of stat(2).
        """
        LOGGER.debug(f"readdir({path})")
        path = Path(path)
        if path == Path("/"):
            for part in self.partitions:
                yield part

    def read(self, path, size, offset, fh):
        """Read content from an object.

        Args:
            path (Path): Path to object.
            size (int): size of content to be read.
            offset (int): starting offset in the file.
            fh: file handle.

        Returns:
            bytes: Raw file content.
        """
        LOGGER.debug(f"read({path}, {size}, {offset})")
        path = Path(path)
        if path.parent == Path("/") and path.name in self.partitions:
            partition = self.partitions[path.name]
            return partition.corrected[offset : offset + size]
        return b""


def mount(image, mount_point, conf):

    if not image.exists():
        LOGGER.warning(f"Image file {image} not found, exiting.")
        return -1

    if not mount_point.exists():
        LOGGER.warning(f"Mount point {mount_point} not found, exiting.")
        return -2

    if not conf.exists():
        LOGGER.warning(f"Configuration file {conf} not found, exiting.")
        return -3
    
    LOGGER.info(f"Mounting corrected image {image} on mount point {mount_point} with configuration {conf}")
    conf = load_config(conf)
    nand = FuseNAND(image, mount_point, conf)
    FUSE(nand, str(mount_point), nothreads=True, foreground=True)
    LOGGER.info(f"Unmounting image {image} from mount point {mount_point}")




import logging
import sys
from argparse import ArgumentParser
from pathlib import Path

from nandtool.config import get_configs
from nandtool.mount import mount
from nandtool.logger import setup_logging

LOGGER = logging.getLogger("nandtool")


if __name__ == "__main__":

    parser = ArgumentParser(description="The parent parser", add_help=False)

    main_parser = ArgumentParser(prog="qnxmount")
    subparsers = main_parser.add_subparsers(title="mount or list", required=True, dest="type")

    parser_mount = subparsers.add_parser("mount", parents=[parser], help="mount (ecc corrected) partitions of the image")
    parser_mount.add_argument("image", type=Path, help="path to image")
    parser_mount.add_argument("-m", "--mount_point", type=Path, help="path to mount point", required=True)
    parser_mount.add_argument("-c", "--config", help="path or key of configuration file", required=True)

    parser_list = subparsers.add_parser("list", parents=[parser], help="list available config files")

    args = main_parser.parse_args()


    
    config_files = get_configs()

    if args.type == "list":
        print("Available configurations:")
        for file in sorted(config_files):
            print(f"  {file}")
        sys.exit(0)

    setup_logging(LOGGER)

    if args.type == "mount":
        if args.config in config_files:
            args.config = Path(config_files[args.config])
        else:
            args.config = Path(args.config)

        sys.exit(mount(args.image, args.mount_point, args.config))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nandtool']

package_data = \
{'': ['*'], 'nandtool': ['configs/*']}

install_requires = \
['bchlib>=0.14.0,<0.15.0',
 'crcmod>=1.7,<2.0',
 'dynaconf>=3.1.11,<4.0.0',
 'fusepy>=3.0.1,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pytest>=7.2.1,<8.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'nandtool',
    'version': '0.1.0',
    'description': 'ECC correction for NAND flash',
    'long_description': '# Nand tool\n\nThis project contains code to do ECC correction for NAND flash. The tool has been created because ECC configurations are varied and can have many options. We want a general tool that can perform ECC corrections on any NAND flash. The configuration of the ECC must be known and should be recorded in a configuration file (`.toml`). \n\nFor each partition in the file system a different ECC structure can be defined and each partition will be mounted on a separate folder of the specified mount point. The mounting is done using fusepy. For a few systems the configuration files are in this repository. If no configuration file is present, the section [Structure of a Configuration File](#Structure-of-a-Configuration-File) explains how a config file should be structured. \n\n\n## Getting started\n\nSet up your Python virtual environment and activate the environment:\n\n```commandline\npython3 -m venv venv\nsource ./venv/bin/activate\n```\n\nInstall the linux dependencies below and install nandtool in the virtual environment:\n\n```commandline\nsudo apt install fuse python3-dev build-essential\npip install nandtool\n```\n\n\n## Usage\n\nTo list the available configurations:\n\n```shell\npython3 -m nandtool list\n```\n\nTo use the tool to mount (an ECC corrected) image:\n\n```shell\npython3 -m nandtool mount /image -m /mountpoint -c /config\n```\n\nIf mounting succeeds you will see the log message `"Mounting image /image on mount point /mountpoint with configuration /config"` appear and the process will hang. Navigate to the given mount point with another terminal session or a file browser to access the NAND partitions.\n\nUnmounting can be done from the terminal with:\n\n```shell\nsudo umount /mountpoint\n```\n\nThe logs will show show that the image was successfully unmounted and nandtool will exit.\n\nNote that the partition is not saved to the mount point. Once you unmount, the partitions are removed. If you want to save the partition binaries, you have to do so by hand.\n\n\n## Structure of a Configuration File\n\nThe [example.toml](config/example.toml) shows what a config file should look like and the different sections and parameters are explained in the comments. Setting up the config requires one to know the partitions in the NAND image. How the partion info can be found is discussed in [the next section](#NAND-Partition-Info). \n\nFor the partitions a layout must be defined and the page size is set. The SIMPLE layout shows the required parameters for a parition with ecc correction, and the COMPLEX layout shows the additional optional parameters that can be set. The NOECC layout shows that it is also possible to extract the user data without performing ecc correction. Note that not all partitions in the image have to be listed. The nandtool will only correct and extract the partitions listed in the config file. \n\nThe ordering of ecc bytes and ecc protected data within a page differs greatly per NAND flash. Therefore, this tool allows the user to set per byte whether it is ecc protected data, user data, or ecc code. This is done using intervals. Moreover, a page may be split up into several chunks (usually four) on which error correction needs to be performed separately. For each chunk the intervals need to be defined. \n\nFor example, in the simple layout, we have four chunks of 512 bytes of ecc protected data at the start of the page. The user data is equal to the ecc protected data in this case. The ecc bytes for each chunk are located at the end of the page at offsets 2050, 2064, 2078, 2092, and each 13 bytes long.\nIn the complex layout the protected data bytes are immidiately followed by the ecc bytes, which is repeated four times. Also, the protected data intervals are different from the user data intervals. \n\nIt is also possible to define configurations that do not perform ECC correction, but simply extract the uncorrected user data for instance. In this case simply omit the parameters that define ECC correction. Some configurations without ECC correction are provided and labeled `noecc`.\n\nFeel free to create a merge request if you create configs for systems not yet available in this repo.\n\n\n## NAND Partition Info\n\nOn QNX systems [dumpifs](https://github.com/FrancisHoogendijk/dumpifs) can be used to get insight into the partition table. First, extract the user data from the image without performing the ECC correction. This can also be done with a config file (the ones called `noecc`):\n\n```commandline\npython3 -m nandtool /image -m /mountpoint -c /config\n```\n\nClone the [dumpifs](https://github.com/FrancisHoogendijk/dumpifs) repository and use the tool in the repo. The image in the command below is the file we just mounted with the command above.\n\n```commandline\ndumpifs -x /image -d /output_dir\n```\n\nIn the output directory look for `etc/system/config/nand_partition.txt`. In this text file the offsets (in blocks) of the partitions are given.\n\nOn systems that do not run QNX, finding the partition table may involve manually delving into the binary for the information.\n\n\n## Contributing\n\nIf you want develop and contribute to the tool, first fork the repository. Contributions can be submitted as a merge request. \n\nTo get started clone the forked repository and create a virtual environment. Install poetry, necessary linux packages, and dependencies.\n\n```commandline\nsudo apt install fuse python3-dev build-essential\npip install .\n```\n\n',
    'author': 'Francis Hoogendijk',
    'author_email': 'f.hoogendijk@nfi.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NetherlandsForensicInstitute/nandtool',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import mmap
import bchlib
import logging
import numpy as np
from itertools import chain
from tqdm import tqdm
from pathlib import Path

LOGGER = logging.getLogger(__name__)


def modify_buffer(buffer, reverse=False, invert=False):
    array = np.frombuffer(buffer, "u1")
    if invert:
        array = array ^ 0xFF
    if reverse:
        array = ((array * 0x0202020202) & 0x010884422010) % 1023  # C trickery to bitwise reverse a byte
    return array.astype("u1").tobytes()


def build_partitions(image, conf):
    partitions = dict()
    for partition in conf.partitions:
        partconf = conf[partition]
        LOGGER.info(f"Start building partition: {partition}")
        partitions[partition] = NAND(image, partconf)
        LOGGER.info(f"Done building partition: {partition}")
    return partitions


class NAND:
    def __init__(self, path, conf):
        self.path = path
        self.f = open(self.path, "rb")
        conf = conf.to_dict()  # If we keep using the dynaconf object throughout the code, access will be slow
        self.conf = conf
        layout = conf["layout"]
        if conf["endblock"] == -1:
            conf["endblock"] = Path(path).stat().st_size // layout["blocksize"] - 1
        num_blocks = conf["endblock"] + 1 - conf["startblock"]
        self.size_with_oob = num_blocks * layout["blocksize"]
        self.mm = mmap.mmap(
            self.f.fileno(), length=self.size_with_oob, offset=conf["startblock"] * layout["blocksize"], access=mmap.ACCESS_READ
        )
        self.pagesize = layout["pagesize"]
        self.oobsize = layout["oobsize"]

        if layout["ecc_algorithm"]:
            self.bch = bchlib.BCH(layout["ecc_algorithm"]["poly"], layout["ecc_algorithm"]["t"])
        else: 
            self.bch = None
        # Avoid diving into dynaconf for every single chunk correction, which is slow!
        self.ecc = layout["ecc"]
        self.ecc_protected_data = layout["ecc_protected_data"]

        self.user_data = layout["user_data"]
        self.left_shift = layout["left_shift_ecc_buf"]
        
        self.ecc_protected_data_reverse = layout["ecc_protected_data_reverse"]
        self.ecc_protected_data_invert = layout["ecc_protected_data_invert"]
        self.ecc_reverse = layout["ecc_reverse"]
        self.ecc_invert = layout["ecc_invert"]

        if hasattr(layout, "etfs") or 'etfs' in layout:
            self.etfs_layout = layout["etfs"]
        else:
            self.etfs_layout = dict()

        pagesize = sum((end - start for chunk in self.user_data for start, end in chunk))
        if self.etfs_layout:
            pagesize += 16
        # pagesize = layout["pagesize"] if not self.etfs_layout else layout["pagesize"] + 16
        self.corrected_partition_size = num_blocks * layout["pages_per_block"] * pagesize

        # This actually starts the ECC correction
        self.corrected_bits = 0
        self.corrected = self.correct_partition()
        LOGGER.info(f"Corrected {self.corrected_bits} bits")

    def __del__(self):
        self.mm.close()
        self.f.close()

    @staticmethod
    def rebuild_buffer(mapping, raw_pagesize=None):
        if not raw_pagesize:
            raise ValueError("Invalid arguments, need raw_pagesize!")
        out = b""
        cursor = 0
        for offset, buf in sorted(mapping.items(), key=lambda x: x[0]):
            if offset > cursor:
                out += b"\xff" * (offset - cursor)
                cursor += offset - cursor
            out += buf
            cursor += len(buf)
        if len(out) < raw_pagesize:
            out += b"\xff" * (raw_pagesize - len(out))
        return out

    def bch_correct_chunk(self, data_ranges, ecc_ranges, data_map):
        data = b"".join(data_map[start] for start, _ in data_ranges)
        ecc = b"".join(data_map[start] for start, _ in ecc_ranges)
        data = modify_buffer(data, invert=self.ecc_protected_data_invert, reverse=self.ecc_protected_data_reverse)
        ecc = modify_buffer(ecc, invert=self.ecc_invert, reverse=self.ecc_reverse)
        leftshift = self.left_shift
        if leftshift == 4:
            ecc = bytes.fromhex(ecc.hex()[1:] + "0")
        flips, data_corrected, ecc_corrected = self.bch.decode(data, ecc)
        if leftshift == 4:
            ecc_corrected = bytes.fromhex("0" + ecc_corrected.hex()[:-1])
        data_corrected = modify_buffer(data_corrected, invert=self.ecc_protected_data_invert, reverse=self.ecc_protected_data_reverse)
        ecc_corrected = modify_buffer(ecc_corrected, invert=self.ecc_invert, reverse=self.ecc_reverse)
        if flips == -1:
            raise ValueError("Uncorrectable number of bitflips.")
        if flips:
            self.corrected_bits += flips
            LOGGER.debug(f"Detected {flips} flips")

        for start, end in data_ranges:
            length = end - start
            data_map[start] = data_corrected[:length]
            data_corrected = data_corrected[length:]
        assert not data_corrected
        for start, end in ecc_ranges:
            length = end - start
            data_map[start] = ecc_corrected[:length]
            ecc_corrected = ecc_corrected[length:]
        assert not ecc_corrected

    def correct_page(self, page):
        # Do correction
        data_map = dict()
        all_erased = True
        for chunk in chain(self.ecc_protected_data, self.ecc):
            for start, end in chunk:
                buf = page[start:end]
                if buf != b"\xff" * len(buf):
                    all_erased = False
                data_map[start] = buf
        if not all_erased:
            for chunk_data_ranges, chunk_ecc_ranges in zip(self.ecc_protected_data, self.ecc):
                self.bch_correct_chunk(chunk_data_ranges, chunk_ecc_ranges, data_map)  # update data_map with corrected data
        # Rebuild page data
        corrected_page = self.rebuild_buffer(data_map, raw_pagesize=self.pagesize + self.oobsize)
        return corrected_page

    def correct_partition(self):
        out = []
        raw_pagesize = self.pagesize + self.oobsize

        for i in tqdm(range(0, self.size_with_oob, raw_pagesize)):
            page = self.mm[i : i + raw_pagesize]
            if self.ecc and self.ecc_protected_data and self.bch:
                corrected_page = self.correct_page(page)
            else:
                corrected_page = page
            userdata = b""
            for chunk in self.user_data:
                for start, end in chunk:
                    userdata += corrected_page[start:end]
            # Append transaction in case of ETFS
            if self.etfs_layout:
                userdata += self.build_transaction(corrected_page)
            out.append(userdata)
        return b"".join(out)

    def build_transaction(self, page):
        transaction = b""
        start, end = self.etfs_layout["fid"]
        transaction += page[start:end]
        transaction += b"\x00" * 2  # fid padding?
        for key in ("cluster", "nclusters"):
            start, end = self.etfs_layout[key]
            transaction += page[start:end]
        transaction += b"\x00" * 2  # tacode and dacode
        start, end = self.etfs_layout["sequence"]
        transaction += page[start:end]
        return transaction
